if (typeof module !== 'undefined' && typeof module.exports !== 'undefined')
    Vector3 = require("./Vector3.js");

/**
 * @class Matrix3
 * @author samiBendou
 * @brief 3x3 matrices
 * @details `Matrix3` class represents dense 3x3 matrices.
 *
 * Components are stored in memory as an aggregation of three Vector3 rows : `x`, `y`, `z`.
 *
 * Access a component of the matrix with the syntax `m.i.j` where `i` and `j` might be equal to `x`, `y` or `z`.
 * eg. `m.x.x`, `m.y.z`, ...
 *
 * Construct a matrix by explicitly giving components.
 * eg. ` new Matrix3(1, 2, 3)` fills the first row `x` with `{x = 1, y = 2, z = 3}`.
 *
 * #### Features
 *
 * - Use extended **vector space operations** and **geometrical operations** from `Vector3`.
 *
 * - Compute  **algebraical operations** such as inversion, product or determinant.
 *
 * - Create matrices using **generators** in a Matlab-style syntax such as `eye`, `ones`, ...
 *
 * - All the generators from `Vector3` are extended to `Matrix3` class.
 *
 * @property x {Vector3} first row
 * @property y {Vector3} second row
 * @property z {Vector3} third row
 */
class Matrix3 {

    constructor(xx = 0, xy = 0, xz = 0,
                yx = 0, yy = 0, yz = 0,
                zx = 0, zy = 0, zz = 0) {

        this.x = new Vector3(xx, xy, xz);
        this.y = new Vector3(yx, yy, yz);
        this.z = new Vector3(zx, zy, zz);
    }

    /**
     * @brief i-th row
     * @param i {number} index of the row to get
     * @return {Vector3} value of the row
     */
    row(i) {
        let labels = ["x", "y", "z"];
        return this[labels[i]].copy();
    }

    /**
     * @brief j-th column
     * @param j {number} index of the column to get
     * @return {Vector3} value of the column
     */
    col(j) {
        let labels = ["x", "y", "z"];
        return new Vector3(this.x[labels[j]], this.y[labels[j]], this.z[labels[j]]);
    }

    fill(s) {
        this.x.fill(s);
        this.y.fill(s);
        this.z.fill(s);
        return this;
    }

    add(m) {
        this.x.add(m.x);
        this.y.add(m.y);
        this.z.add(m.z);
        return this;
    }

    sub(m) {
        this.x.sub(m.x);
        this.y.sub(m.y);
        this.z.sub(m.z);
        return this;
    }

    get opp() {
        this.x.opp;
        this.y.opp;
        this.z.opp;
        return this;
    }

    mul(s) {
        this.x.mul(s);
        this.y.mul(s);
        this.z.mul(s);
        return this;
    }

    div(s) {
        this.x.div(s);
        this.y.div(s);
        this.z.div(s);
        return this;
    }

    /**
     * @brief transpose the matrix
     * @returns {Matrix3} value of the transposed matrix
     */
    get trans() {
        let copy = this.copy();

        copy.x.y = this.y.x;
        copy.x.z = this.z.x;

        copy.y.x = this.x.y;
        copy.y.z = this.z.y;

        copy.z.x = this.x.z;
        copy.z.y = this.y.z;

        return copy;
    }

    /**
     * @brief product between two matrix
     * @param m {Matrix3} matrix to multiply
     * @returns {Matrix3} value of the product
     */
    prod(m) {
        let mTrs = m.copy().trans;
        let copy = this.copy();

        copy.x.x = this.x.scal(mTrs.x);
        copy.x.y = this.x.scal(mTrs.y);
        copy.x.z = this.x.scal(mTrs.z);

        copy.y.x = this.y.scal(mTrs.x);
        copy.y.y = this.y.scal(mTrs.y);
        copy.y.z = this.y.scal(mTrs.z);

        copy.z.x = this.z.scal(mTrs.x);
        copy.z.y = this.z.scal(mTrs.y);
        copy.z.z = this.z.scal(mTrs.z);

        return copy;
    }

    /**
     * @brief product between matrix and vector
     * @details Also called linear mapping between matrix and vector.
     * @param u {Vector3} vector to multiply
     * @returns {Vector3} value of linear mapping
     */
    map(u) {
        let copy = u.copy();

        copy.x = this.x.scal(u);
        copy.y = this.y.scal(u);
        copy.z = this.z.scal(u);

        return copy;
    }

    /**
     * @brief determinant of a matrix
     * @returns {number} value of the determinant
     */
    get det() {
        return this.x.x * this.y.y * this.z.z + this.x.y * this.y.z * this.z.x + this.x.z * this.y.x * this.z.y
            - this.x.z * this.y.y * this.z.x - this.x.y * this.y.x * this.z.z - this.x.x * this.y.z * this.z.y;
    }

    /**
     * @brief inverse of a matrix
     * @details returns a `NaN` matrix when the determinant is zero.
     * @returns {Matrix3} value of the inverse matrix
     */
    get inv() {
        return (new Matrix3(this.y.y * this.z.z - this.y.z * this.z.y,
            this.x.z * this.z.y - this.x.y * this.z.z,
            this.x.y * this.y.z - this.x.z * this.y.y,

            this.y.z * this.z.x - this.y.x * this.z.z,
            this.x.x * this.z.z - this.x.z * this.z.x,
            this.x.z * this.y.x - this.x.x * this.y.z,

            this.y.x * this.z.y - this.y.y * this.z.x,
            this.x.y * this.z.x - this.x.x * this.z.y,
            this.x.x * this.y.y - this.x.y * this.y.x)).div(this.det);
    }

    copy() {
        let copy = new Matrix3();

        copy.x = this.x.copy();
        copy.y = this.y.copy();
        copy.z = this.z.copy();

        return copy;
    }

    isEqual(m) {
        return this.x.isEqual(m.x) && this.y.isEqual(m.y) && this.z.isEqual(m.z);
    }

    isZero() {
        return this.x.isZero() && this.y.isZero() && this.z.isZero();
    }

    toString() {
        return this.x.toString() + "\n" + this.y.toString() + "\n" + this.z.toString();
    }

    /**
     * @brief transform matrix to array
     * @return {Array} bi-dimensional array containing rows of the matrix
     */
    toArray() {
        return [this.x.toArray(), this.y.toArray(), this.z.toArray()];
    }

    static get zeros() {
        return new Matrix3();
    }

    static get ones() {
        return new Matrix3(1, 1, 1, 1, 1, 1, 1, 1, 1);
    }

    /**
     * @brief identity matrix
     * @details Diagonal matrix filled with `1`.
     * @returns {Matrix3} value of identity matrix
     */
    static get eye() {
        return new Matrix3(1, 0, 0, 0, 1, 0, 0, 0, 1);
    }

    /**
     * @brief scalar matrix
     * @details Diagonal matrix filled with a single value.
     * @param s {number} scalar value
     * @returns {Matrix3} value of scalar matrix
     */
    static scal(s) {
        return new Matrix3(s, 0, 0, 0, s, 0, 0, 0, s);
    }

    /**
     * @brief canonical matrix
     * @details Matrix with `0` everywhere except in `i`, `j` position where there is a `1`.
     * @param i {number} row index of `1`
     * @param j {number} column index of `1`
     * @returns {Matrix3} value of canonical matrix
     */
    static e(i, j) {
        let labels = ["x", "y", "z"];
        let can = Matrix3.zeros;

        can[labels[i]][labels[j]] = 1;
        return can;
    }

    static sum(matrices) {
        return matrices.reduce(function (prev, cur) {
            return prev.copy().add(cur);
        });
    }

    static comb(scalars, matrices) {
        return matrices.reduce(function (prev, cur, index) {
            return prev.copy().mul(scalars[index - 1]).add(cur.copy().mul(scalars[index]));
        });
    }

    /**
     * @details product of the matrices in array
     * @details The product is performed in the same order as the elements in the array.
     * @param matrices {Array} array of `Vector3`
     * @returns {Matrix3} value of the product
     */
    static prod(matrices) {
        return matrices.reduce(function (prev, cur) {
            return prev.copy().prod(cur);
        });
    }

    /**
     * @brief rotation matrix of axis (`0`, `ex`)
     * @details Anticlockwise rotation.
     * @param theta {number} angle of rotation
     * @returns {Matrix3} value of rotation matrix for specified angle
     */
    static rotX(theta) {
        return new Matrix3(1, 0, 0,
            0, Math.cos(theta), -Math.sin(theta),
            0, Math.sin(theta), Math.cos(theta));
    }

    /**
     * @brief rotation matrix of axis (`0`, `ey`)
     * @details Anticlockwise rotation.
     * @param theta {number} angle of rotation
     * @returns {Matrix3} value of rotation matrix for specified angle
     */
    static rotY(theta) {
        return new Matrix3(Math.cos(theta), 0, Math.sin(theta),
            0, 1, 0,
            -Math.sin(theta), 0, Math.cos(theta));
    }

    /**
     * @brief rotation matrix of axis (`0`, `ez`)
     * @details Anticlockwise rotation.
     * @param theta {number} angle of rotation
     * @returns {Matrix3} value of rotation matrix for specified angle
     */
    static rotZ(theta) {
        return new Matrix3(Math.cos(theta), -Math.sin(theta), 0,
            Math.sin(theta), Math.cos(theta), 0,
            0, 0, 1);
    }

    /**
     * @brief rotation matrix with around axis
     * @details Anticlockwise rotation.
     * @returns {function(number):Matrix3} function that generates rotation `Matrix3` for given angle
     */
    static makeRot(axis) {
        //R = P + cos(theta) * (I - P) + sin(theta) * Q

        let id = Matrix3.eye; // antisymmetric representation of u
        let q = Matrix3.zeros;
        let u = axis.copy().div(axis.r); // normalized axis
        let p = Matrix3.tens(u); // projection on rotation axis
        let r = p.copy();

        q.x = u.cross(Vector3.ex);
        q.y = u.cross(Vector3.ey);
        q.z = u.cross(Vector3.ez);
        q = q.trans;

        return function (theta) {
            return r.add(id.sub(p).mul(Math.cos(theta))).add(q.mul(Math.sin(theta)));
        };
    }

    /**
     * @brief affine transformation of the vector
     * @param m {Matrix3} matrix of the transformation
     * @param v {Vector3} vector of the transformation
     * @returns {function(Vector3): Vector3} function that computes affine transform for given `Vector3`
     */
    static makeAffine(m, v) {
        return function (u) {
            return m.map(u).add(v)
        };
    }

    /**
     * @brief tensor product of two vectors
     * @details Tensor product is the matrix obtained from two vectors such that `m.i.j = u.i * v.j`.
     * @param u {Vector3} first vector to transform
     * @param v {Vector3} second vector to transform
     * @returns {Matrix3} value of the tensor product
     */
    static tens(u, v) {
        let right = v || u;
        return new Matrix3(
            u.x * right.x, u.x * right.y, u.x * right.z,
            u.y * right.x, u.y * right.y, u.y * right.z,
            u.z * right.x, u.z * right.y, u.z * right.z,
        );
    }

    /**
     * @brief creates a matrix with given array
     *  @details The order of the rows in matrix is the same as in `arr` array
     * @param arr {Array} bi-dimensional array containing rows of the matrix
     * @returns {Matrix3} new instance of matrix
     */
    static fromArray(arr) {
        return new Matrix3(
            arr[0][0], arr[0][1], arr[0][2],
            arr[1][0], arr[1][1], arr[1][2],
            arr[2][0], arr[2][1], arr[2][2]);
    }
}

if (typeof module !== 'undefined' && typeof module.exports !== 'undefined')
    module.exports = Matrix3;
else
    window.Matrix3 = Matrix3;